from dnsdist_console import Console
from dnsdist_console import Statistics
import unittest

import dns.resolver

                           
my_resolver = dns.resolver.Resolver(configure=False)
my_resolver.nameservers = ['127.0.0.1']
my_resolver.port = 5553

r = my_resolver.resolve('www.github.com', 'a')
try:
    r = my_resolver.resolve('12345678', 'a')
except dns.resolver.NXDOMAIN:
    pass

class TestStatistic(unittest.TestCase):
    def setUp(self):
        self.console = Console(host="127.0.0.1", port=5199,
                               key="GQpEpQoIuzA6kzgwDokX9JcXPXFvO1Emg1wAXToJ0ag=")

    def tearDown(self):
        self.console.disconnect()
        
    def test_get_global(self):
        """get global stats"""
        s = Statistics(console=self.console)

        print(s["global"])

        self.assertGreater(int(s["global"]["queries"]), 2)

    def test_get_topqueries(self):
        """get top queries"""
        s = Statistics(console=self.console)

        print(s["top-queries"])

        self.assertEqual(len(s["top-queries"]), 2)
        self.assertGreater(int(s["top-queries"][0]["hits"]), 0)

    def test_get_nxdomain(self):
        """get nxdomain"""
        s = Statistics(console=self.console)

        print(s["top-nxdomain"])

        self.assertEqual(len(s["top-nxdomain"]), 1)
        self.assertGreater(int(s["top-nxdomain"][0]["hits"]), 0)

    def test_get_topclients(self):
        """get top clients"""
        s = Statistics(console=self.console)

        print(s["top-clients"])

        self.assertEqual(len(s["top-clients"]), 1)
        self.assertGreater(int(s["top-clients"][0]["hits"]), 0)

    def test_backends(self):
        """backends"""
        s = Statistics(console=self.console)

        # 4 backends ?
        self.assertEqual(len(s["backends"]), 4)

        tc = { 
                "0": {'Name': 'a long name with spa', 'Pools': 'a pool with spaces'},
                "1": {'Name': 'name with spaces', 'Pools': 'apoolwithoutspaces'},
                "2": {'Name': 'namewithoutspaces', 'Pools': 'shortpool'},
                "3":  {'Name': 'and a long long long', 'Pools': 'a long long long long pool'}
            }
        for b in s["backends"]:
            if b["#"] in tc:
                self.assertEqual(b["Name"], tc[b["#"]]["Name"])
                self.assertEqual(b["Pools"], tc[b["#"]]["Pools"])


from dnsdist_console import Console
import unittest

class TestConnect(unittest.TestCase):
    def setUp(self):
        self.console = Console(host="127.0.0.1", port=5199,
                               key="GQpEpQoIuzA6kzgwDokX9JcXPXFvO1Emg1wAXToJ0ag=")
        
    def tearDown(self):
        self.console.disconnect()
        
    def test_show_version(self):
        """connect and get version"""             
        
        o = self.console.send_command(cmd="showVersion()")
        self.assertRegex(o, ".*dnsdist.*")


class TestConnectV6(unittest.TestCase):
    def setUp(self):
        self.console = Console(host="::1", port=5199,
                               key="GQpEpQoIuzA6kzgwDokX9JcXPXFvO1Emg1wAXToJ0ag=")

    def tearDown(self):
        self.console.disconnect()

    def test_show_version(self):
        """connect and get version"""

        o = self.console.send_command(cmd="showVersion()")
        self.assertRegex(o, ".*dnsdist.*")




import libnacl
import libnacl.utils
import base64

class Key:
    def generate(self):
        """generate console key"""
        k = libnacl.utils.salsa_key()
        return base64.b64encode(k).decode("utf8")
        




from dnsdist_console import Statistics

import time
import sys

class Dashboard:
    def __init__(self, console):
        """statistics class"""
        self.console = console
     
        self.show()
        
    def show(self):
        """show dashboard updated every seconds"""
        prev_queries = 0
        prev_cpu_sys = 0
        prev_cpu_user = 0
                
        lines = {
                    "Uptime (seconds)": "--",
                    "Number of queries": "--",
                    "Query per second": "--",
                    "ACL drops": "--",
                    "Dynamic drops": "--",
                    "Rule drops": "--",
                    "CPU Usage (%s)": "--",
                    "Cache hitrate": "--"
                }

        while True:
            try:
                # get stats from dnsdist
                stats = Statistics(console=self.console)
                global_stats = stats["global"]
                
                qps = int(global_stats["queries"]) - prev_queries
                prev_queries = int(global_stats["queries"])
                cpu = (int(global_stats["cpu-sys-msec"])+int(global_stats["cpu-user-msec"]) - prev_cpu_sys - prev_cpu_user) / 10
                prev_cpu_sys = int(global_stats["cpu-sys-msec"])
                prev_cpu_user = int(global_stats["cpu-user-msec"])
            
                lines["Uptime (seconds)"] = global_stats["uptime"]
                lines["Number of queries"] = global_stats["queries"]
                lines["Query per second"] = qps
                lines["CPU Usage (%s)"] = cpu
                lines["ACL drops"] = global_stats["acl-drops"]
                lines["Rule drops"] = global_stats["rule-drop"]
                lines["Cache hitrate"] = global_stats["cache-hits"]
                lines["Dynamic drops"] = global_stats["dyn-blocked"]

                # reprint the lines    
                sys.stdout.write("\033[1mDashboard for dnsdist\033[0m\n")
                sys.stdout.write("\n")
                sys.stdout.write("Global:\n")
                for k,v in lines.items():
                    sys.stdout.write("\t%s: %s\n" % (k,v))
                sys.stdout.write("Backends:\n")
                for s in stats["backends"]:
                    if not len(s["name"]):
                        s["name"] = "--"
                    if not len(s["pools"]):
                        s["pools"] = "--"
                    sys.stdout.write("\t#%s / %s / %s / %s\n" % (s["#"],s["address"],s["name"],s["pools"]) )
                    sys.stdout.write("\t\tNumber of queries: %s\n" % s["queries"])
                    sys.stdout.write("\t\tQuery per second: %s\n" % s["qps"])
                    sys.stdout.write("\t\tNumber of drops: %s\n" % s["drops"])
                sys.stdout.write("\n")
                sys.stdout.write("Ctrl+C to exit\n")
        
                time.sleep(1)
                
                
                # move up cursor and delete whole line
                sys.stdout.write("\x1b[1A\x1b[2K") 
                sys.stdout.write("\x1b[1A\x1b[2K")
                sys.stdout.write("\x1b[1A\x1b[2K")
                for k,v in lines.items():
                    sys.stdout.write("\x1b[1A\x1b[2K") 
                sys.stdout.write("\x1b[1A\x1b[2K")
                for s in stats["backends"]:
                    sys.stdout.write("\x1b[1A\x1b[2K")
                    sys.stdout.write("\x1b[1A\x1b[2K") 
                    sys.stdout.write("\x1b[1A\x1b[2K") 
                    sys.stdout.write("\x1b[1A\x1b[2K")
                sys.stdout.write("\x1b[1A\x1b[2K")
                sys.stdout.write("\x1b[1A\x1b[2K")
                    
                del stats
            except KeyboardInterrupt:
                break


from dnsdist_console import HashPassword
import unittest

class TestHashPassword(unittest.TestCase):
    def test_generate_hashpassword(self):
        """generate hash password"""             
        h = HashPassword().generate(password="superpassword")
        self.assertRegex(h, "^\$scrypt\$.*")


import base64
import secrets
import scrypt

class HashPassword:
    def generate(self, password, work_factor=10, parallel_factor=1, block_size=8, salt_size=16, output_size=32, pwhash_prefix="$scrypt$"):
        """generate hash password for web server"""
        result = []
        result.append(pwhash_prefix)

        result.append("ln=")
        result.append("%s" % work_factor)

        result.append(",p=")
        result.append("%s" % parallel_factor)

        result.append(",r=")
        result.append("%s" % block_size)

        result.append("$")
        salt = secrets.token_bytes(salt_size)
        result.append("%s" % base64.b64encode(salt).decode())
        result.append("$")

        out = scrypt.hash(password, salt, N=2**work_factor, p=parallel_factor, r=block_size, buflen=output_size)
        result.append("%s" % base64.b64encode(out).decode())

        return "".join(result)


import socket
import libnacl
import libnacl.utils
import struct
import base64
import time
from dnsdist_console import util

class Console:
    def __init__(self, key, host="127.0.0.1", port=5199):
        """authenticator class"""
        self.console_host = host
        self.console_port = port
        self.console_key = base64.b64decode(key) 
        
        self.nonce_c = libnacl.utils.rand_nonce()
        self.nonce_s = None
        self.nonce_w = None
        self.nonce_r = None
        self.sock = None
        self.sock_timeout = 3.0  # Increased from 1.0 to be more tolerant
        
        self.connect_to()
        
    def encrypt_command(self, cmd, nonce):
        """encrypt console"""
        cmd = cmd.encode('utf-8')
        return libnacl.crypto_secretbox(cmd, nonce, self.console_key)
        
    def decrypt_response(self, data, nonce):
        """decrypt console"""
        result = libnacl.crypto_secretbox_open(data, nonce, self.console_key)
        return result.decode('utf-8')
        
    def incremente_nonce(self, nonce):
        """incremente nonce"""
        v = int.from_bytes(nonce[:4], "big")
        v += 1
        return v.to_bytes(4, byteorder='big') + nonce[4:]
        
    def disconnect(self):
        """disconnect"""
        if self.sock is not None:
            try:
                self.sock.close()
            except:
                pass
            self.sock = None
        
    def connect_to(self):
        """connect to console"""
        # prepare socket
        if util.is_ipv4_address(self.console_host):
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.sock.settimeout(self.sock_timeout)
        
        try:
            # connect to the server
            self.sock.connect((self.console_host, self.console_port))
            # send our nonce
            self.sock.send(self.nonce_c)
            
            # waiting to receive server nonce
            self.nonce_s = self.sock.recv(len(self.nonce_c))
            if len(self.nonce_s) != len(self.nonce_c):
                raise Exception("incorrect nonce size")
      
            # init reading and writing nonce
            half_nonce = int(len(self.nonce_c) / 2)
            self.nonce_r = self.nonce_c[0:half_nonce] + self.nonce_s[half_nonce:]
            self.nonce_w = self.nonce_s[0:half_nonce] + self.nonce_c[half_nonce:]
            # send empty command to check if the hanshake is ok
            try:
                  self.send_command(cmd="")
            except Exception as e:
                  raise Exception("hanshake error: %s" % e)
        except Exception as e:
            self.disconnect()
            raise Exception(f"Connection error: {e}")
            
    def reconnect_if_needed(self):
        """Check connection and reconnect if necessary"""
        if self.sock is None:
            self.connect_to()
            return True
        return False
            
    def send_command(self, cmd, max_retries=2):
        """send command to console and return output with auto-reconnect"""
        retries = 0
        last_error = None
        
        while retries <= max_retries:
            try:
                # Try to reconnect if socket is None
                self.reconnect_if_needed()
                
                # encrypt command
                encrypted_cmd = self.encrypt_command(cmd, self.nonce_w)
                
                # send data size header
                self.sock.send(struct.pack("!I", len(encrypted_cmd)))
                # send encrypted command
                self.sock.send(encrypted_cmd)
                
                # waiting to receive data size
                data = self.sock.recv(4)
                if not data:
                    raise Exception("no response size received")
                    
                # unpack response size
                (response_size,) = struct.unpack("!I", data)
                # waiting to response response according to the response size
                data = self.sock.recv(response_size)
                while len(data) < response_size:
                    data += self.sock.recv(response_size - len(data))
                # decrypt data
                r = self.decrypt_response(data, self.nonce_r)
                # incremente nonce for next command
                self.nonce_r = self.incremente_nonce(nonce=self.nonce_r)
                self.nonce_w = self.incremente_nonce(nonce=self.nonce_w)
                # return response output
                return r
            except (BrokenPipeError, socket.error) as e:
                # Socket related error - attempt reconnection
                last_error = e
                retries += 1
                self.disconnect()
                if retries <= max_retries:
                    time.sleep(1)  # Small delay before retry
                    # Will reconnect on next loop iteration
            except Exception as e:
                # Other errors, just raise
                raise Exception(f"Command error: {e}")
                
        # If we get here, we've exhausted our retries
        raise Exception(f"Failed after {max_retries} retries: {last_error}")



from dnsdist_console.console import Console
from dnsdist_console.key import Key
from dnsdist_console.statistics import Statistics
from dnsdist_console.dashboard import Dashboard
from dnsdist_console.hashpwd import HashPassword


from dnsdist_console import Key
import unittest

class TestConnect(unittest.TestCase):
    def test_generate_key(self):
        """generate key in base64"""             
        k = Key().generate()
        self.assertRegex(k, "[a-zA-Z0-9+\/]+={0,2}")


#!/usr/bin/python

import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()
    
KEYWORDS = ('dnsdist console client')

setuptools.setup(
    name="dnsdist_console",
    version="1.6.0",
    author="Denis MACHARD",
    author_email="d.machard@gmail.com",
    description="Python client for dnsdist console",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/dmachard/dnsdist_console",
    packages=['dnsdist_console', 'tests'],
    include_package_data=True,
    platforms='any',
    keywords=KEYWORDS,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=[
        "libnacl",
        "scrypt",
    ]
)



import re

class Statistics:
    def __init__(self, console, top_num=10):
        """statistics class"""
        self.c = console
        self.top_num = top_num
        self.__stats__ = {"global": {}, "backends": [] }
        
        self.fetch()
        
    def __repr__(self):
        """print stats"""
        return "%s" % self.__stats__
        
    def __getitem__(self, key):
        """get stats"""
        return self.__stats__[key]
        
    def fetch(self):
        """Get statistics from dnsdist"""
        
        # dump all stats
        o = self.c.send_command(cmd="dumpStats()")

        for l in o.splitlines():
              r = re.findall("\S+", l)
              for i in range(0, len(r), 2):
                   self.__stats__["global"][r[i]] = r[i+1]
                 
        # dump servers stats        
        o = self.c.send_command(cmd="showServers()")
        o_list = o.splitlines()[:-1]

        svr_hdr = o_list[0]
        
        svr_stats = []
        headers = re.sub('\s+', ' ', svr_hdr).split(' ')

        # Validate expectations about fields
        if 'Name' not in headers:
            raise ValueError("Missing Name field in dnsdist showServers()")
        if headers[0] != '#':
            raise ValueError("Unexpected first header in dnsdist showServers()")

        for s in o_list[1:]:
            # Find out the server name, expect it at pos 0
            server_id = s.split(' ')[0]
            srv_name_o = self.c.send_command(cmd="getServer("+server_id+")")
            srv_name = srv_name_o.rstrip()
            # The name output by showServers() is limited to 20 chars
            srv_name_trunc = srv_name[:20]
            # Sub with a copy that removes spaces to allow spliting the other fields
            srv_name_nowsp = srv_name_trunc.replace(' ','_')
            s = s.replace(srv_name_trunc, srv_name_nowsp)
            # Split, except the last fields (Pools), which contains spaces too
            svr = dict(zip(headers, re.sub('\s+', ' ', s).split(' ', len(headers) - 1)))
            # Restore name as output by showServers()
            svr['Name'] = srv_name_trunc
            svr_stats.append(svr)
            
        self.__stats__["backends"] = svr_stats
        
        self.__stats__["top-queries"] = self.top_queries(num=self.top_num)
        
        self.__stats__["top-nxdomain"] = self.top_nxdomain(num=self.top_num)
        
        self.__stats__["top-clients"] = self.top_clients(num=self.top_num)
        
    def top_queries(self, num=10):
        """get top queries"""
        o = self.c.send_command(cmd="topQueries(%s)" % num)
        return self.decode_top_output(o=o)
        
    def top_nxdomain(self, num=10):
        """get top non-existent domain"""
        o = self.c.send_command(cmd="topResponses(%s,DNSRCode.NXDOMAIN)" % num)
        return self.decode_top_output(o=o)
        
    def top_clients(self, num=10):
        """get top clients"""
        o = self.c.send_command(cmd="topClients(%s)" % num)
        return self.decode_top_output(o=o)
        
    def decode_top_output(self, o):
        """decode top output"""
        o_list = []
        for l in o.splitlines()[:-1]:
            r = re.findall("\S+", l)
            o_list.append({"entry": r[1], "hits": r[2]})
        return o_list